#!/usr/bin/env python3
"""generate boxplots from descriptive stats in yaml"""

import os
import sys
import typing as T

import yaml

Data = dict[str, T.Any]

PLOT_SPECIFICS = {
    'hints': {
        'ylabel': '{latency [us]}',
        'ymax': '300',
        'xmax': '3',
        'xtick': '{1,2}',
        'xticklabels': '{pipe, pipe-no-hints}',
        'title': 'Wakeup Hints',
    },
    'latency': {
        'ymax': '1500',
        'xtick': '{1,2,3}',
        'xticklabels': '{pipe-steal-nc, pipe, pipe-nc}',
        'title': 'Pipe Latency Variants'
    },
    'contention': {
        'ymax': '1100',
        'xtick': '{1,2,3}',
        'xticklabels': '{no-sleep, vanilla}',
        'title': 'Global queue contention'
    },
    'l-vs-ll': {
        'ymax': '1500',
        'xtick': '{1,2,3,4}',
        'xticklabels':
        '{ll-n-nc, n-nc\\vertPhantom{}, ll-s-nc, s-nc\\vertPhantom{}}',
        'title': ' Lockless vs Locked'
    },
    'n-s-c': {
        'ylabel': '{latency [us]}',
        'ymax': '1300',
        'xtick': '{1,2,3,4}',
        'xticklabels': '{n, n-nc, s, s-nc}',
        'title': 'Additional Completer'
    },
    'n-vs-ns': {
        'ylabel': '{latency [us]}',
        'ymax': '150',
        'xtick': '{1,2,3}',
        'xticklabels': '{n-nc\\vertPhantom{}, s-ns\\vertPhantom{}, ll-s-ns}',
        'title': 'Notifications vs No Sleep'
    },
    'strategies': {
        'ymax': '250',
        'xtick': '{1,2,3}',
        'xticklabels': '{vanilla, pipe, \\waitfd{}}',
        'title': 'Sleep Strategies'
    },
}


def generate_plot_specifics() -> str:
    """Generate plot specifics for the plot passed via the environment"""
    plot = os.environ.get('BOXPLOT', None)
    if not plot:
        return ''

    if plot not in PLOT_SPECIFICS:
        print(f'ERROR: specified plot "{plot}" unknown', file=sys.stderr)
        sys.exit(1)

    tikz_lines = []
    for key, value in PLOT_SPECIFICS[plot].items():
        tikz_lines.append(f'{key}={value}')

    return ',\n'.join(tikz_lines)


def parse_data(stream) -> T.Optional[Data]:
    """Convert yamls string to dict"""
    return yaml.safe_load(stream)

TIKZ_TEMPLATE = \
"""\\documentclass[]{standalone}
\\input{common.tex}
\\begin{document}
\\begin{tikzpicture}
\\begin{axis}[
    mark=x,
    boxplot/draw direction=y,
    legend columns=2,
    area legend,
    xmin=0,
    xtick pos=bottom,
    ytick pos=left,
    xminorticks=false,
    yminorticks=false,
    ymajorgrids=true,
    legend style={
        draw=none,
        fill=none,
        at={(0.5,-0.1)},
        anchor=north
    },
    $PLOT_SPECIFICS$
]
$PLOTS$
\\end{axis}
\\end{tikzpicture}
\\end{document}
"""

PLOT_TEMPLATE = \
"""\\addplot+[
    mark=x,
    boxplot prepared={{
          lower whisker={lower_whisker},
          lower quartile={lower_quartile},
          median={median},
          average={mean},
          upper quartile={upper_quartile},
          upper whisker={upper_whisker},
      }},
] coordinates{{{outlier_cords}}};
%\\addlegendentry{{{variant}}}
"""


def generate_boxplot(data: Data) -> str:
    """generate a tikzboxplot"""

    plots = ''

    for variant, variant_stats in data.items():
        stats = variant_stats['times']
        outlier_cords = ''
        for outlier in stats['outliers']:
            outlier_cords += f'(0, {outlier}) '
        plots += PLOT_TEMPLATE.format(variant=variant.replace('_', '-'),
                                      outlier_cords=outlier_cords,
                                      **stats)

    tikz = TIKZ_TEMPLATE.replace('$PLOTS$', plots)

    plot_specifics = generate_plot_specifics()
    tikz = tikz.replace('$PLOT_SPECIFICS$', plot_specifics)
    return tikz


def main():
    """read yaml and print generated boxplots"""
    if len(sys.argv) == 1:
        data = parse_data(sys.stdin)
    else:
        with open(sys.argv[1], 'r', encoding='utf-8') as yaml_file:
            data = parse_data(yaml_file)

    if data is None:
        print('No data to plot', file=sys.stderr)
        sys.exit(1)

    tikz = generate_boxplot(data)

    if len(sys.argv) > 2:
        with open(sys.argv[2], 'w', encoding='utf-8') as saveto:
            print(tikz, file=saveto)
    else:
        print(tikz)


if __name__ == '__main__':
    main()
