#!/usr/bin/env python3
"""Collect and summarize evaluation results"""

import argparse
import sys
import typing as T
from pathlib import Path

import numpy
import yaml

Variant = str
Measurements = list[int]
VariantData = dict[str, T.Union[Measurements, float]]
Data = dict[Variant, VariantData]


def collect(result_dir) -> Data:
    """Collect the data in result_dir and return calculated averages"""
    result_dir = Path(result_dir)

    data: Data = {}
    for result_file_path in result_dir.iterdir():
        variant = result_file_path.name.split('.')[0]
        variant_data = data.get(variant, {})

        if result_file_path.suffix == '.out':
            with open(result_file_path, 'r', encoding='utf-8') as result_file:
                # record times in us
                variant_data['times'] = [
                    int(line) // 1000 for line in result_file.readlines()
                ]

            data[variant] = variant_data

        if result_file_path.suffixes == ['.energy', '.stats']:
            with open(result_file_path, 'r', encoding='utf-8') as result_file:
                # record times in us
                variant_data.update({
                    line.split(';')[2]:
                    float(line.split(';')[0])
                    for line in result_file.readlines()[2:]
                })

            data[variant] = variant_data

    return data


Avgs = dict[Variant, dict[str, float]]


def calc_avgs(data) -> Avgs:
    """Calculate only averages from data"""
    avgs = {
        t: {
            k: numpy.mean(ms)
            for k, ms in d.items()
        }
        for t, d in data.items()
    }

    return avgs


Outliers = list[int]
DescriptiveStats = dict[str, T.Union[float, Outliers]]
VariantStats = dict[str, DescriptiveStats]
Stats = dict[Variant, VariantStats]


def calc_stats(data: Data) -> Stats:
    """Calculate and return descriptive stats of all measurements in data"""
    stats = {}
    for variant, variant_data in data.items():
        variant_stats: VariantStats = {}
        stats[variant] = variant_stats

        for key, measurements in variant_data.items():
            key_stats: DescriptiveStats = {}
            variant_stats[key] = key_stats

            # single data point
            if isinstance(measurements, float):
                key_stats['mean'] = measurements
                continue

            key_stats['mean'] = numpy.mean(measurements)
            key_stats['std'] = numpy.std(measurements)

            measurements.sort()
            key_stats['min'] = measurements[0]
            key_stats['max'] = measurements[-1]
            key_stats['median'] = float(numpy.median(measurements))
            upper_quartile = float(numpy.percentile(measurements, 75))
            key_stats['upper_quartile'] = upper_quartile
            lower_quartile = float(numpy.percentile(measurements, 25))
            key_stats['lower_quartile'] = lower_quartile
            iqr = upper_quartile - lower_quartile

            outliers = []
            # find whiskers
            i = 0
            while measurements[i] < lower_quartile - 1.5 * iqr:
                i += 1
            key_stats['lower_whisker'] = measurements[i]
            outliers = measurements[:i]

            i = len(measurements) - 1
            while measurements[i] > upper_quartile + 1.5 * iqr:
                i -= 1
            key_stats['upper_whisker'] = measurements[i]
            outliers += measurements[i + 1:]
            key_stats['outliers'] = outliers

            # convert everything to float to easily dump it using pyyaml
            for stat, value in key_stats.items():
                if isinstance(value, list):
                    continue
                key_stats[stat] = float(value)
    return stats


def summarize(avgs: T.Optional[Avgs] = None,
              stats: T.Optional[Stats] = None,
              keys: T.Optional[T.Sequence[str]] = None,
              desc_stats: T.Optional[T.Sequence[str]] = None) -> int:
    """Print a summary for each selected key of the collected stats"""
    data = avgs or stats

    if data is None:
        print('no data to summarize')
        return 1

    for variant in data:
        if avgs:
            print(f'{variant}: {avgs[variant]}')
        else:
            assert stats is not None
            for key in keys or stats[variant].keys():
                for stat in desc_stats or stats[variant][key].keys():
                    print(
                        f'{variant}-{key}-{stat}: {stats[variant][key][stat]}')

    return 0


def main():
    """program entry point"""
    parser = argparse.ArgumentParser()
    parser.add_argument('-s',
                        '--desc-stats',
                        help='print all stats not only means',
                        nargs='*')
    parser.add_argument("-i",
                        "--implementations",
                        help="implementations to plot",
                        nargs='+')
    parser.add_argument('--yaml',
                        help='dump statistics as yaml',
                        action='store_true')
    parser.add_argument('--dataref',
                        help='dump statistics as dataref',
                        action='store_true')
    parser.add_argument('result_dir',
                        help='directory containing the results to summarize')

    args = parser.parse_args()

    data = collect(args.result_dir)
    stats = calc_stats(data)
    if args.implementations:
        stats = {v: s for v, s in stats.items() if v in args.implementations}

    if args.yaml:
        print(yaml.safe_dump(stats))
        sys.exit(0)

    if args.dataref:
        for impl, var_stats in stats.items():
            for key, desc_stats in var_stats.items():
                for stat, val in desc_stats.items():
                    if stat == 'outliers':
                        continue
                    print(rf'\drefset{{{impl}/{key}/{stat}}}{{{val}}}')
        sys.exit(0)

    sys.exit(summarize(stats=stats, desc_stats=args.desc_stats))


if __name__ == '__main__':
    main()
